/*
 * Copyright © 2021 Christian Heitzmann, SimplexaCode AG
 * www.simplexacode.ch
 * Version 2021.02
 */
package ch.simplexacode.completablefuture;

import java.io.IOException;
import java.math.BigInteger;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadLocalRandom;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

public class CompletableFutureDemo {

  private static final int RUN_COUNT = 10;
  private static final int NUMBER_COUNT_MAX = 10_000_000;
  private static final int NUMBER_BIT_COUNT = 128;
  private static final int CPU_POOL_THREAD_COUNT = 8;
  private static final String FILE_NAME_PREFIX = "numbers_";
  private static final String FILE_NAME_SUFFIX = ".txt";

  private static final String TABLE_HEADER =
      "CrtNs | RemEv | RemPr | RemSq | WrtFi | PrAvg | PrMed | PrMin | PrMax | Result";
  private static final Pattern COMMON_POOL_WORKER_REGEX_PATTERN =
      Pattern.compile("ForkJoinPool\\.commonPool-worker-(\\d{1,2})");
  private static final Pattern POOL_THREAD_REGEX_PATTERN =
      Pattern.compile("pool-(\\d)-thread-(\\d{1,2})");
  private final Lock printLock = new ReentrantLock();
  private int printedLineCount = 0;

  public static void main(String[] args) {
    CompletableFutureDemo demo = new CompletableFutureDemo();
    demo.executeSequentiallyOnMainThread();
    demo.executeCompletableFuturesOnCommonPool();
    demo.executeCompletableFuturesOnDifferentThreadPools();
    demo.executeCompletableFuturesWithSplitting();
  }

  private void executeSequentiallyOnMainThread() {
    for (int i = 1; i <= RUN_COUNT; i++) {
      Set<BigInteger> numberSet = createRandomNumberSet();
      removeEvenNumbers(numberSet);
      removePrimeNumbers(numberSet);
      removeSquareNumbers(numberSet);
      writeNumbersToFile(numberSet);
      printRoughAverage(numberSet);
      printRoughMedian(numberSet);
      printMin(numberSet);
      printMax(numberSet);
    }
  }

  private void executeCompletableFuturesOnCommonPool() {
    @SuppressWarnings("unchecked")
    CompletableFuture<Void>[] completableFutureArray = new CompletableFuture[RUN_COUNT];
    for (int runNumber = 0; runNumber < RUN_COUNT; runNumber++) {
      CompletableFuture<Void> completableFuture =
          CompletableFuture.supplyAsync(this::createRandomNumberSet)
              .thenApplyAsync(this::removeEvenNumbers)
              .thenApplyAsync(this::removePrimeNumbers)
              .thenApplyAsync(this::removeSquareNumbers)
              .thenApplyAsync(this::writeNumbersToFile)
              .thenApplyAsync(this::printRoughAverage)
              .thenApplyAsync(this::printRoughMedian)
              .thenApplyAsync(this::printMin)
              .thenAcceptAsync(this::printMax);
      completableFutureArray[runNumber] = completableFuture;
    }
    CompletableFuture.allOf(completableFutureArray).join();
  }

  private void executeCompletableFuturesOnDifferentThreadPools() {
    ExecutorService cpuService = Executors.newFixedThreadPool(CPU_POOL_THREAD_COUNT);
    ExecutorService ioService = Executors.newSingleThreadExecutor();
    @SuppressWarnings("unchecked")
    CompletableFuture<Void>[] completableFutureArray = new CompletableFuture[RUN_COUNT];
    for (int runNumber = 0; runNumber < RUN_COUNT; runNumber++) {
      CompletableFuture<Void> completableFuture =
          CompletableFuture.supplyAsync(this::createRandomNumberSet, cpuService)
              .thenApplyAsync(this::removeEvenNumbers, cpuService)
              .thenApplyAsync(this::removePrimeNumbers, cpuService)
              .thenApplyAsync(this::removeSquareNumbers, cpuService)
              .thenApplyAsync(this::writeNumbersToFile, ioService)
              .thenApplyAsync(this::printRoughAverage, cpuService)
              .thenApplyAsync(this::printRoughMedian, cpuService)
              .thenApplyAsync(this::printMin, cpuService)
              .thenAcceptAsync(this::printMax, cpuService);
      completableFutureArray[runNumber] = completableFuture;
    }
    CompletableFuture.allOf(completableFutureArray).join();
    cpuService.shutdown();
    ioService.shutdown();
  }

  private void executeCompletableFuturesWithSplitting() {
    ExecutorService cpuService = Executors.newFixedThreadPool(CPU_POOL_THREAD_COUNT);
    ExecutorService ioService = Executors.newSingleThreadExecutor();
    @SuppressWarnings("unchecked")
    CompletableFuture<Void>[] completableFutureArray = new CompletableFuture[RUN_COUNT];
    for (int runNumber = 0; runNumber < RUN_COUNT; runNumber++) {
      CompletableFuture<Set<BigInteger>> part1 =
          CompletableFuture.supplyAsync(this::createRandomNumberSet, cpuService)
              .thenApplyAsync(this::removeEvenNumbers, cpuService)
              .thenApplyAsync(this::removePrimeNumbers, cpuService)
              .thenApplyAsync(this::removeSquareNumbers, cpuService);
      CompletableFuture<Void> part2a = part1.thenAcceptAsync(this::writeNumbersToFile, ioService);
      CompletableFuture<Void> part2b = part1.thenAcceptAsync(this::printRoughAverage, cpuService);
      CompletableFuture<Void> part2c = part1.thenAcceptAsync(this::printRoughMedian, cpuService);
      CompletableFuture<Void> part2d = part1.thenAcceptAsync(this::printMin, cpuService);
      CompletableFuture<Void> part2e = part1.thenAcceptAsync(this::printMax, cpuService);
      CompletableFuture<Void> part2 =
          CompletableFuture.allOf(part2a, part2b, part2c, part2d, part2e);
      completableFutureArray[runNumber] = part2;
    }
    CompletableFuture.allOf(completableFutureArray).join();
    cpuService.shutdown();
    ioService.shutdown();
  }

  private Set<BigInteger> createRandomNumberSet() {
    printCurrentThreadName(0);
    Set<BigInteger> numberSet = new HashSet<>();
    ThreadLocalRandom random = ThreadLocalRandom.current();
    int numberCount = random.nextInt(NUMBER_COUNT_MAX) + 1;
    for (int i = 1; i <= numberCount; i++) {
      BigInteger number = new BigInteger(NUMBER_BIT_COUNT, random);
      numberSet.add(number);
    }

    return numberSet;
  }

  private Set<BigInteger> removeEvenNumbers(Set<BigInteger> numberSet) {
    printCurrentThreadName(1);
    numberSet.removeIf(n -> n.mod(BigInteger.TWO).equals(BigInteger.ZERO));

    return numberSet;
  }

  private Set<BigInteger> removePrimeNumbers(Set<BigInteger> numberSet) {
    printCurrentThreadName(2);
    numberSet.removeIf(n -> n.isProbablePrime(10));

    return numberSet;
  }

  private Set<BigInteger> removeSquareNumbers(Set<BigInteger> numberSet) {
    printCurrentThreadName(3);
    numberSet.removeIf(n -> n.sqrtAndRemainder()[1].equals(BigInteger.ZERO));

    return numberSet;
  }

  private Set<BigInteger> writeNumbersToFile(Set<BigInteger> numberSet) {
    printCurrentThreadName(4);
    int randomFileNumber = ThreadLocalRandom.current().nextInt(1000, 10_000);
    Path path = Paths.get(FILE_NAME_PREFIX + randomFileNumber + FILE_NAME_SUFFIX);
    Set<String> numberStringSet =
        numberSet.stream().map(BigInteger::toString).collect(Collectors.toSet());
    try {
      Files.write(path, numberStringSet);
    } catch (IOException e) {
      e.printStackTrace();
    }

    return numberSet;
  }

  private Set<BigInteger> printRoughAverage(Set<BigInteger> numberSet) {
    printCurrentThreadName(5);
    BigInteger sum = numberSet.stream().reduce(BigInteger::add).orElse(BigInteger.ZERO);
    BigInteger count = BigInteger.valueOf(numberSet.size());
    BigInteger average = sum.divide(count);
    printResult("Average", average);

    return numberSet;
  }

  private Set<BigInteger> printRoughMedian(Set<BigInteger> numberSet) {
    printCurrentThreadName(6);
    BigInteger median =
        numberSet.stream().sorted().skip(numberSet.size() / 2).findFirst().orElse(BigInteger.ZERO);
    printResult(" Median", median);

    return numberSet;
  }

  private Set<BigInteger> printMin(Set<BigInteger> numberSet) {
    printCurrentThreadName(7);
    BigInteger min = numberSet.stream().min(BigInteger::compareTo).orElse(BigInteger.ZERO);
    printResult("Minimum", min);

    return numberSet;
  }

  private void printMax(Set<BigInteger> numberSet) {
    printCurrentThreadName(8);
    BigInteger max = numberSet.stream().max(BigInteger::compareTo).orElse(BigInteger.ZERO);
    printResult("Maximum", max);
  }

  private void printCurrentThreadName(int indentation) {
    printLock.lock();
    try {
      if (printedLineCount % 10 == 0) {
        System.out.println(TABLE_HEADER);
      }
      for (int i = 1; i <= indentation; i++) {
        System.out.print("        ");
      }
      String threadName = Thread.currentThread().getName();
      Matcher commonPoolWorkerMatcher = COMMON_POOL_WORKER_REGEX_PATTERN.matcher(threadName);
      Matcher poolThreadMatcher = POOL_THREAD_REGEX_PATTERN.matcher(threadName);
      if (commonPoolWorkerMatcher.find()) {
        String workerNumberString = commonPoolWorkerMatcher.group(1);
        System.out.format("cp-%2s%n", workerNumberString);
      } else if (poolThreadMatcher.find()) {
        String poolNumberString = poolThreadMatcher.group(1);
        String threadNumberString = poolThreadMatcher.group(2);
        System.out.format("%2s-%2s%n", poolNumberString, threadNumberString);
      } else {
        System.out.format("%5s%n", threadName);
      }
      printedLineCount++;
    } finally {
      printLock.unlock();
    }
  }

  private void printResult(String resultName, BigInteger result) {
    printLock.lock();
    try {
      if (printedLineCount % 10 == 0) {
        System.out.println(TABLE_HEADER);
      }
      for (int i = 1; i <= 9; i++) {
        System.out.print("        ");
      }
      System.out.format("%s: %s%n", resultName, result);
      printedLineCount++;
    } finally {
      printLock.unlock();
    }
  }
}
