/*
 * Copyright © 2021 Christian Heitzmann, SimplexaCode AG
 * www.simplexacode.ch
 * Version 2021.02
 */
package ch.simplexacode.completablefuture;

import java.math.BigInteger;
import java.util.concurrent.CancellationException;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

public class FutureDemo {

  private static final String NUMBER_STRING = "74747474747474747";
  private static final int TIMEOUT = 15;

  public static void main(String[] args) {
    FutureDemo demo = new FutureDemo();
    demo.execute();
  }

  private void execute() {
    BigInteger number = new BigInteger(NUMBER_STRING);

    ExecutorService service = Executors.newCachedThreadPool();
    Future<Boolean> primeFuture = service.submit(() -> numberIsPrime(number));
    service.submit(() -> waitBusilyAndPrintResult(primeFuture));
    service.submit(() -> waitBlockinglyAndPrintResult(primeFuture));

    service.shutdown();
  }

  private boolean numberIsPrime(BigInteger number) {
    BigInteger divisorMax = number.sqrt();
    for (BigInteger divisor = BigInteger.TWO;
        divisor.compareTo(divisorMax) <= 0;
        divisor = divisor.add(BigInteger.ONE)) {
      if (Thread.currentThread().isInterrupted()) {
        return false;
      }
      if (number.divideAndRemainder(divisor)[1].equals(BigInteger.ZERO)) {
        return false;
      }
    }

    return true;
  }

  private void waitBusilyAndPrintResult(Future<Boolean> future) {
    while (!future.isDone()) {
      System.out.println("Waiting busily ...");
      try {
        Thread.sleep(1000);
      } catch (InterruptedException e) {
        System.err.println("Wait was interrupted.");
        Thread.currentThread().interrupt();
      }
    }
    try {
      boolean result = future.get();
      System.out.format("%s is %sa prime number.%n", NUMBER_STRING, result ? "" : "not ");
    } catch (InterruptedException e) {
      System.err.println("Thread was interrupted.");
      Thread.currentThread().interrupt();
    } catch (ExecutionException e) {
      System.err.println("Computation threw an exception:");
      e.printStackTrace();
    } catch (CancellationException e) { // RuntimeException
      System.err.println("Computation was canceled.");
    }
  }

  private void waitBlockinglyAndPrintResult(Future<Boolean> future) {
    System.out.println("Waiting blockingly ...");
    try {
      boolean result = future.get(TIMEOUT, TimeUnit.SECONDS);
      System.out.format("%s is %sa prime number.%n", NUMBER_STRING, result ? "" : "not ");
    } catch (InterruptedException e) {
      System.err.println("Thread was interrupted.");
      Thread.currentThread().interrupt();
    } catch (ExecutionException e) {
      System.err.println("Computation threw an exception:");
      e.printStackTrace();
    } catch (TimeoutException e) {
      System.err.println("Wait for result timed out.");
      System.err.println("Computation will now be canceled.");
      future.cancel(true);
    } catch (CancellationException e) { // RuntimeException
      System.err.println("Computation was canceled.");
    }
  }
}
